//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "Wrapper.h"
#include "StringHelp.h"
#include <shellapi.h>
#include <math.h>
#include <IniFiles.hpp>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TWrapperForm *WrapperForm;
//---------------------------------------------------------------------------
__fastcall TWrapperForm::TWrapperForm(TComponent* Owner)
    : TForm(Owner)
{
}
//---------------------------------------------------------------------------
//updated 15 May 2022

void __fastcall TWrapperForm::FormCreate(TObject *Sender)
{
    //Open Persistant Data & initialise variables
    TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

    Edit3->Text = IniFile->ReadString("GCodeWrapper","MaxSegmentLength","1.0");
    Edit4->Text = IniFile->ReadString("GCodeWrapper","MaxAngle","10");
    Edit5->Text = IniFile->ReadString("GCodeWrapper","ZAxisCentre","0.0");
    Edit6->Text = IniFile->ReadString("GCodeWrapper","WorkSurface","7.5");
    Edit7->Text = IniFile->ReadString("GCodeWrapper","AAxisCentreZ","0.0");
    OpenDialog1->InitialDir = IniFile->ReadString("GCodeWrapper","InitialDir",ExtractFileDir(Application->ExeName));
    ComboBox1->ItemIndex = IniFile->ReadInteger("GCodeWrapper","Axis",1);
    ComboBox2->ItemIndex = IniFile->ReadInteger("GCodeWrapper","RotaryAxis",0);

    delete IniFile;

}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::FormDestroy(TObject *Sender)
{
    //Save Persistant Data
    TIniFile *IniFile = new TIniFile(ExtractFileDir(Application->ExeName) + "\\Data.ini");

    IniFile->WriteString("GCodeWrapper","MaxSegmentLength",Edit3->Text);
    IniFile->WriteString("GCodeWrapper","MaxAngle",Edit4->Text);
    IniFile->WriteString("GCodeWrapper","ZAxisCentre",Edit5->Text);
    IniFile->WriteString("GCodeWrapper","WorkSurface",Edit6->Text);
    IniFile->WriteString("GCodeWrapper","AAxisCentreZ",Edit7->Text);
    IniFile->WriteString("GCodeWrapper","InitialDir",OpenDialog1->InitialDir);
    IniFile->WriteInteger("GCodeWrapper","Axis",ComboBox1->ItemIndex);
    IniFile->WriteInteger("GCodeWrapper","RotaryAxis",ComboBox2->ItemIndex);

    delete IniFile;
    
}
//---------------------------------------------------------------------------

void __fastcall TWrapperForm::OriginalGCodeButtonClick(TObject *Sender)
{
    if (OpenDialog1->Execute()) {
        Edit1->Text = OpenDialog1->FileName;
        OpenDialog1->InitialDir = ExtractFilePath(Edit1->Text);
        Edit2->Text = ChangeFileExt(Edit1->Text, " Wrapped.nc");
        SaveDialog1->InitialDir = ExtractFilePath(Edit1->Text);
        }

}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::NewGCodeButtonClick(TObject *Sender)
{
    if (SaveDialog1->Execute()) {
        Edit2->Text = SaveDialog1->FileName;
        SaveDialog1->InitialDir = ExtractFilePath(Edit2->Text);
        }

    
}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::WrapButtonClick(TObject *Sender)
{
    if (!Edit1->Text.Length() || !Edit2->Text.Length()) {
        Application->MessageBox("Need two file names\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (Edit1->Text.LowerCase() == Edit2->Text.LowerCase()) {
        Application->MessageBox("Both file names the same\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (!Edit3->Text.Length()) {
        Application->MessageBox("Please enter a maximum value for\nthe segment length of an arc\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (!Edit4->Text.Length()) {
        Application->MessageBox("Please enter a maximum value for\nthe angle subtended by a segment around an arc\n\nConversion Aborted.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (ToDoubleDef(Edit5->Text,0.0) == ToDoubleDef(Edit6->Text,7.5)) {
        Application->MessageBox("The nominal work Z is the same height as the axis,\nso the nominal work perimiter is zero.","Warning", MB_OK | MB_ICONERROR);
        Label3->Caption = "Conversion Aborted";
        return;
        }

    if (ToDoubleDef(Edit5->Text,0.0) > ToDoubleDef(Edit6->Text,7.5)) {
        Application->MessageBox("The nominal work Z is below the axis, so the work is hollowed out and will\nbe mirrored about the axis. This may be a problem.","Warning", MB_OK | MB_ICONERROR);
        }

    if (FileExists(Edit1->Text) && !FileExists(Edit2->Text)) {
        try {
            WrapButton->Enabled = false;
            OriginalGCodeButton->Enabled = false;
            NewGCodeButton->Enabled = false;
            HelpButton->Enabled = false;
            CloseButton->Enabled = false;
            Edit1->Enabled = false;
            Edit2->Enabled = false;
            Edit3->Enabled = false;
            Edit4->Enabled = false;
            Refresh();

            TStringList *GCodeInputFile = new TStringList();
            TStringList *GCodeOutputFile = new TStringList();
            Label3->Caption = "Opening File";
            Refresh();
            GCodeInputFile->LoadFromFile(Edit1->Text);


            //This conditioning is needed to change the text into codes separated by spaces, problem highlighted by Ken Moffat
            Label3->Caption = "Checking the file";
            Refresh();
            AnsiString temp;
            AnsiString tempout;
            int pointer;
            char c,c1;
            for (int i = 0; i < GCodeInputFile->Count; i++) {
                pointer = 0;
                temp = GCodeInputFile->Strings[i];
                tempout = "";
                for (int j = 1; j < temp.Length(); j++) {
                    c = temp[j];
                    c1 = temp[j+1];
                    tempout = tempout + AnsiString(c);
                    if (((c > 47 && c < 58) || c == 46) && ((c1 > 64 && c1 < 91) || (c1 > 96 && c1 < 123))) {
                        tempout = tempout + " ";
                        }
                    }
                tempout = tempout + AnsiString(c1);
                GCodeInputFile->Strings[i] = tempout;
                }

            //second set of conditioning needed to ensure that there is no detrimental effect from not repeating the G code every line.
            AnsiString LastGCode = "G00";
            AnsiString Number;
            for (int i = 0; i < GCodeInputFile->Count; i++) {
                pointer = 0;
                temp = GCodeInputFile->Strings[i].UpperCase();
                if (temp.Pos("G")) {
                    //update G code to current one
                    while (temp.Length() && LeftString(temp,1) != "G") {
                        temp = ChopOffFirstChars(temp,1);
                        }
                    if (temp.Length()) LastGCode = ExtractWord(temp," ");
                    }
                if (!temp.Pos("G") && (temp.Pos("X") || temp.Pos("Y") || temp.Pos("Z"))) {
                    if (temp.Pos("N")) {
                        Number = ExtractWord(temp,"N") + "N";
                        temp = DeleteWord(temp,"N");
                        Number = Number + ExtractWord(temp," ");
                        temp = DeleteWord(temp," ");
                        } else {
                        Number = "";
                        }
                    tempout = Number + " " + LastGCode + " " + temp;
                    GCodeInputFile->Strings[i] = tempout;
                    }
                }





            Label3->Caption = "Running through the file";
            Refresh();

            //do not need, or really use, all these variables, but it may be useful in the future to have them
            int counter = 0;
            //AnsiString temp;
            //AnsiString tempout;
            AnsiString Instruction;
            double currentX = 999.0;
            double currentY = 999.0;
            double currentZ = 999.0;
            double currentA = 999.0;
            double currentB = 999.0;
            double currentC = 999.0;
            //use of currentMachineX allows the axes to be mapped as required while keeping the arc segment generation working
            double currentMachineX = 999.0;
            double currentFeedrate = 100.0;
            //no idea what the rapid feedrate is, so use my machine's rapid X/Y feedrate. Z is slower
            double rapidFeedrate = 2300.0;
            double currentRetractLevel = 0.0;
            TPlane currentPlane = TXYPlane;
            TSpindleDir currentSpindleDir = TCW;
            TCoolant currentCoolant = TNone;
            TUnits Units = TMetric;
            double codeno;
            double currentTime = 0.0;
            int currentTool = 0;
            bool relativeTranslation = false;
            bool relativeCentres = false;
            double targetarclength = ToDoubleDef(Edit3->Text,0.1);
            //max angle in radians
            double maxangle = ToDoubleDef(Edit4->Text,1.0)*M_PI/180.0;
            //cutter compensation
            TCutterComp cutterCompensation = TCCNone;
            TPathControl pathControl = TVelocity;

            bool unknowncode = false;
            bool firstmove = true;

            //Initialising wrapping variables
            TAxis Axis = (TAxis) ComboBox1->ItemIndex;
            TAxis MappedAxis = (TAxis) (ComboBox2->ItemIndex + 3);
            double InputAxisZ = ToDoubleDef(Edit5->Text,0.0);
            double NominalWorkSurfaceZ = ToDoubleDef(Edit6->Text,7.5);
            double OutputAxisZ = ToDoubleDef(Edit7->Text,0.0);
            double mmPerDegree = M_PI * (NominalWorkSurfaceZ - InputAxisZ) / 180.0;
            if (Axis == TYAxis) mmPerDegree = -mmPerDegree;

            DynamicArray<TErrorType> Errors;

            //for injected lines
            int injectedcounter = 0;


            //Write file header
            GCodeOutputFile->Add("(G Code Wrapped by the Chestnut Pens GCodeWrapper Program.)");
            GCodeOutputFile->Add("(Toolpath Creation Program  R.W.Williams, 2013)");
            GCodeOutputFile->Add("(Max Straight segment length = " + DP(4,targetarclength) + "mm)");
            GCodeOutputFile->Add("(Max arc angle of segment around centre = " + DP(4,maxangle * 180.0/M_PI) + "degrees)");
            GCodeOutputFile->Add("(Note: All dimensions are converted to mm)");
            GCodeOutputFile->Add("(Original file Axis height Z = " + DP(4,InputAxisZ) + ")");
            GCodeOutputFile->Add("(Original file Nominal Work height Z = " + DP(4,NominalWorkSurfaceZ) + ")");
            GCodeOutputFile->Add("(This file Axis height Z = " + DP(4,OutputAxisZ) + ")");
            switch (Axis) {
                case TXAxis:    {
                                GCodeOutputFile->Add("(X axis is wrapped...)");
                                }
                                break;
                case TYAxis:    {
                                GCodeOutputFile->Add("(Y axis is wrapped...)");
                                }
                                break;
                default:        {
                                GCodeOutputFile->Add("(Unknown axis is wrapped. Program error.)");
                                }
                                break;
                }
            switch (MappedAxis) {
                case TAAxis:    {
                                GCodeOutputFile->Add("(...onto the A axis)");
                                }
                                break;
                case TBAxis:    {
                                GCodeOutputFile->Add("(...onto the B axis)");
                                }
                                break;
                case TCAxis:    {
                                GCodeOutputFile->Add("(...onto the C axis)");
                                }
                                break;
                default:        {
                                GCodeOutputFile->Add("(...onto an unknown destination axis. Program error.)");
                                }
                                break;
                }
            //Go through the file now
            while (counter < GCodeInputFile->Count) {
                if (!(counter % 1000)) {
                    Label3->Caption = "Line " + AnsiString(counter) + " of " + AnsiString(GCodeInputFile->Count);
                    if (!(counter %10000)) {
                        Refresh();
                        } else {
                        Label3->Refresh();
                        }
                    } // end of  'if (!(counter % 1000))'
                temp = GCodeInputFile->Strings[counter];
                tempout = "";

                while (temp.Length()) {
                    temp = temp.Trim() + " ";
                    Instruction = LeftString(temp,1);

                    if (Instruction == "(") {
                        //we have a comment up to the close bracket
                        tempout += LeftString(temp,temp.Pos(")"));
                        temp = DeleteWord(temp,")").Trim();
                        } else {
                        temp = temp.UpperCase();
                        //not a comment, so needs to be dealt with
                        //debug
                        //GCodeOutputFile->Add("(Line = " + temp + ")");
                        temp = DeleteWord(temp,Instruction);
                        codeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                        temp = DeleteWord(temp," ");

                        //now we can find the meaning of the code etc
                        if (Instruction == "D") {
                            //It's a Tool Offset command. No need to keep track of anything related to this.
                            tempout = tempout + "D" + AnsiString(codeno) + " ";
                            }
                        if (Instruction == "F") {
                            //It's a Feed rate instruction
                            currentFeedrate = codeno;
                            if (Units == TInch) currentFeedrate = currentFeedrate * 25.4;
                            tempout = tempout + "F" + AnsiString(codeno) + " ";
                            //if this feedrate is greater than the defined rapid,
                            //set this as rapid.
                            if (codeno > rapidFeedrate) rapidFeedrate = codeno;
                            }

                        if (Instruction == "G") {
                            //It's a G Code, see which one

                            switch ((int)codeno) {

                                case 1:
                                case 0:     {
                                            //Rapid Traverse G0
                                            //Normal Traverse G1
                                            //possible codes to follow: A, B, C, X, Y, Z, G53, F
                                            double feedrate;
                                            if (codeno == 0.0) {
                                                feedrate = rapidFeedrate;
                                                } else {
                                                feedrate = currentFeedrate;
                                                }
                                            double newA = currentA;
                                            double newB = currentB;
                                            double newC = currentC;
                                            double newX = currentX;
                                            double newY = currentY;
                                            double newZ = currentZ;
                                            double newMachineX = currentMachineX;
                                            bool G53found = false;
                                            bool feedrateset = false;

                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z" ||
                                                    LeftString(temp,1) == "F" ||
                                                    LeftString(temp,3) == "G53"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    //ChopOffFirstChars(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    double scale;
                                                    //injected lines are in mm, other lines may be in inches
                                                    if (Units == TInch && injectedcounter == 0) {
                                                        scale = 25.4;
                                                        } else {
                                                        scale = 1.0;
                                                        if (injectedcounter) injectedcounter--;
                                                        }
                                                    if (relativeTranslation) {
                                                            if (localInstruction == "A") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TAAxisInstruction;
                                                                }
                                                            if (localInstruction == "B") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TBAxisInstruction;
                                                                }
                                                            if (localInstruction == "C") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TCAxisInstruction;
                                                                }
                                                            if (localInstruction == "X") {
                                                                switch (Axis) {
                                                                    case TXAxis:    {
                                                                                    switch (MappedAxis) {
                                                                                        case TAAxis:    {
                                                                                                        newA = currentA + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TBAxis:    {
                                                                                                        newB = currentB + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TCAxis:    {
                                                                                                        newC = currentC + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        }
                                                                                    //iteration 5
                                                                                    newX = currentX + localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    case TYAxis:    {
                                                                                    //iteration 5
                                                                                    newX = currentX + localcodeno * scale;
                                                                                    newMachineX = currentMachineX + localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    default:        break;
                                                                    }
                                                                }
                                                            //if (localInstruction == "Y") newY = currentY + localcodeno * scale;
                                                            if (localInstruction == "Y") {
                                                                switch (Axis) {
                                                                    case TXAxis:    {
                                                                                    //iteration 5
                                                                                    newY = currentY + localcodeno * scale;
                                                                                    newMachineX = currentMachineX + localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    case TYAxis:    {
                                                                                    switch (MappedAxis) {
                                                                                        case TAAxis:    {
                                                                                                        newA = currentA + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TBAxis:    {
                                                                                                        newB = currentB + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TCAxis:    {
                                                                                                        newC = currentC + localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        }
                                                                                    //now need to set the Y axis to keep it on track
                                                                                    newY = currentY + localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    default:        break;
                                                                    }
                                                                }
                                                            if (localInstruction == "Z") newZ = currentZ + localcodeno * scale + OutputAxisZ - InputAxisZ;
                                                            } else {
                                                            //Absolute movements are being used
                                                            if (localInstruction == "A") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TAAxisInstruction;
                                                                }
                                                            if (localInstruction == "B") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TBAxisInstruction;
                                                                }
                                                            if (localInstruction == "C") {
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TCAxisInstruction;
                                                                }
                                                            //if (localInstruction == "X") newX = localcodeno * scale;
                                                            //if (localInstruction == "Y") newY = localcodeno * scale;
                                                            if (localInstruction == "X") {
                                                                switch (Axis) {
                                                                    case TXAxis:    {
                                                                                    switch (MappedAxis) {
                                                                                        case TAAxis:    {
                                                                                                        newA = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TBAxis:    {
                                                                                                        newB = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TCAxis:    {
                                                                                                        newC = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        }
                                                                                    //iteration 5
                                                                                    newX = localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    case TYAxis:    {
                                                                                    //iteration 5
                                                                                    newX = localcodeno * scale;
                                                                                    newMachineX = localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    default:        break;
                                                                    }
                                                                }
                                                            //if (localInstruction == "Y") newY = currentY + localcodeno * scale;
                                                            if (localInstruction == "Y") {
                                                                switch (Axis) {
                                                                    case TXAxis:    {
                                                                                    //iteration 5
                                                                                    newY = localcodeno * scale;
                                                                                    newMachineX = localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    case TYAxis:    {
                                                                                    switch (MappedAxis) {
                                                                                        case TAAxis:    {
                                                                                                        newA = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TBAxis:    {
                                                                                                        newB = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        case TCAxis:    {
                                                                                                        newC = localcodeno * scale / mmPerDegree;
                                                                                                        }
                                                                                                        break;
                                                                                        }
                                                                                    //iteration 5
                                                                                    newY = localcodeno * scale;
                                                                                    }
                                                                                    break;
                                                                    default:        break;
                                                                    }
                                                                }
                                                            if (localInstruction == "Z") newZ = localcodeno * scale + OutputAxisZ - InputAxisZ;
                                                            }
                                                    if (localInstruction == "F") {
                                                        feedrate = localcodeno * scale;
                                                        feedrateset = true;
                                                        }
                                                    if (localInstruction == "G") {
                                                        G53found = true;
                                                        }
                                                    temp = temp.Trim();
                                                    }//end of while statement

                                            //calculate distance to travel
                                            double dist = sqrt((newX - currentX) * (newX - currentX)
                                                            + (newY - currentY) * (newY - currentY)
                                                            + (newZ - currentZ) * (newZ - currentZ)
                                                            );
                                            double arc = sqrt((newA - currentA) * (newA - currentA)
                                                            + (newB - currentB) * (newB - currentB)
                                                            + (newC - currentC) * (newC - currentC)
                                                            );

                                            //claculate time taken to travel distance
                                            //note: feedrate is mm per min
                                            double localtime = dist * 60.0/feedrate;
                                            if (arc > dist) localtime = arc * 60.0/feedrate;
                                            if (firstmove) {
                                                firstmove = false;
                                                } else {
                                                currentTime += localtime;
                                                }

                                            //write code to output
                                            tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            if (feedrateset) tempout = tempout + "F" + AnsiString(feedrate) + " ";
                                            if (G53found) tempout = tempout + "G53 ";
                                            if (relativeTranslation) {
                                                /*
                                                if (newA != currentA) tempout = tempout + "A" + DP(4,newA - currentA) + " ";
                                                if (newB != currentB) tempout = tempout + "B" + DP(4,newB - currentB) + " ";
                                                if (newC != currentC) tempout = tempout + "C" + DP(4,newC - currentC) + " ";
                                                */
                                                //if (newX != currentX) tempout = tempout + "X" + DP(4,newX - currentX) + " ";
                                                //if (newY != currentY) tempout = tempout + "Y" + DP(4,newY - currentY) + " ";
                                                //iteration 5
                                                if (newMachineX != currentMachineX) tempout = tempout + "X" + DP(4,newMachineX - currentMachineX) + " ";
                                                if (newZ != currentZ) tempout = tempout + "Z" + DP(4,newZ - currentZ) + " ";
                                                if (newA != currentA) tempout = tempout + "A" + DP(4,newA - currentA) + " ";
                                                if (newB != currentB) tempout = tempout + "B" + DP(4,newB - currentB) + " ";
                                                if (newC != currentC) tempout = tempout + "C" + DP(4,newC - currentC) + " ";
                                                } else {
                                                /*
                                                if (newA != currentA) tempout = tempout + "A" + DP(4,newA) + " ";
                                                if (newB != currentB) tempout = tempout + "B" + DP(4,newB) + " ";
                                                if (newC != currentC) tempout = tempout + "C" + DP(4,newC) + " ";
                                                */
                                                //if (newX != currentX) tempout = tempout + "X" + DP(4,newX) + " ";
                                                //if (newY != currentY) tempout = tempout + "Y" + DP(4,newY) + " ";
                                                //iteration 5
                                                if (newMachineX != currentMachineX) tempout = tempout + "X" + DP(4,newMachineX) + " ";
                                                if (newZ != currentZ) tempout = tempout + "Z" + DP(4,newZ) + " ";
                                                if (newA != currentA) tempout = tempout + "A" + DP(4,newA) + " ";
                                                if (newB != currentB) tempout = tempout + "B" + DP(4,newB) + " ";
                                                if (newC != currentC) tempout = tempout + "C" + DP(4,newC) + " ";
                                                }

                                            //update position of tool & feedrate
                                            currentA = newA;
                                            currentB = newB;
                                            currentC = newC;
                                            currentX = newX;
                                            currentY = newY;
                                            currentZ = newZ;
                                            currentMachineX = newMachineX;
                                            if (codeno == 1.0 && currentFeedrate != feedrate && feedrateset) {
                                                currentFeedrate = feedrate;
                                                }
                                            if (feedrate > rapidFeedrate) rapidFeedrate = feedrate;

                                            } //end of G0/G1
                                            break;
                                case 2:
                                case 3:     {
                                            TSpindleDir Arcdir;
                                            if (codeno == 3.0) {
                                                Arcdir = TCCW;
                                                } else {
                                                Arcdir = TCW;
                                                }
                                            double feedrate = currentFeedrate;
                                            double newA = currentA;
                                            double newB = currentB;
                                            double newC = currentC;
                                            double newX = currentX;
                                            double newY = currentY;
                                            double newZ = currentZ;
                                            double newI = currentX;
                                            double newJ = currentY;
                                            double newK = currentZ;
                                            bool G53found = false;
                                            bool feedrateset = false;
                                            //debuggery
                                            //GCodeOutputFile->Add("(" + temp + ")");
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z" ||
                                                    LeftString(temp,1) == "F" ||
                                                    LeftString(temp,1) == "I" ||
                                                    LeftString(temp,1) == "J" ||
                                                    LeftString(temp,1) == "K" ||
                                                    LeftString(temp,3) == "G53"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    //ChopOffFirstChars(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    double scale;
                                                    if (Units == TInch) {
                                                        scale = 25.4;
                                                        } else {
                                                        scale = 1.0;
                                                        }
                                                    if (relativeTranslation) {
                                                            if (localInstruction == "A") newA = currentA + localcodeno * scale;
                                                            if (localInstruction == "B") newB = currentB + localcodeno * scale;
                                                            if (localInstruction == "C") newC = currentC + localcodeno * scale;
                                                            if (localInstruction == "X") newX = currentX + localcodeno * scale;
                                                            if (localInstruction == "Y") newY = currentY + localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = currentZ + localcodeno * scale;
                                                            } else {
                                                            if (localInstruction == "A") newA = localcodeno * scale;
                                                            if (localInstruction == "B") newB = localcodeno * scale;
                                                            if (localInstruction == "C") newC = localcodeno * scale;
                                                            if (localInstruction == "X") newX = localcodeno * scale;
                                                            if (localInstruction == "Y") newY = localcodeno * scale;
                                                            if (localInstruction == "Z") newZ = localcodeno * scale;
                                                            }
                                                    if (relativeCentres) {
                                                        if (localInstruction == "I") newI += localcodeno * scale;
                                                        if (localInstruction == "J") newJ += localcodeno * scale;
                                                        if (localInstruction == "K") newK += localcodeno * scale;
                                                        } else {
                                                        if (localInstruction == "I") newI = localcodeno * scale;
                                                        if (localInstruction == "J") newJ = localcodeno * scale;
                                                        if (localInstruction == "K") newK = localcodeno * scale;
                                                        }
                                                    if (localInstruction == "F") {
                                                        feedrate = localcodeno * scale;
                                                        feedrateset = true;
                                                        }
                                                    if (localInstruction == "G") {
                                                        G53found = true;
                                                        }
                                                    temp = temp.Trim();
                                                    }//end of while statement
                                            //OK, all arc information has been read in.
                                            switch (currentPlane) {
                                                case TXYPlane:  {
                                                                //right, the arc is in the XY plane
                                                                //create straight lines & inject them into the input file so that they are then converted
                                                                //to the other axis as necessary
                                                                double radius;
                                                                double startangle;
                                                                double endangle;
                                                                double arclength;
                                                                double stepangle;
                                                                int numberofsteps;
                                                                int numberofarcsteps;
                                                                radius = sqrt(pow(currentX - newI,2) + pow(currentY - newJ,2));
                                                                int localcounter = counter;
                                                                AnsiString localtemp;
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((currentX-newI)/radius > 1.0) {
                                                                    startangle = acos(1.0);
                                                                    } else {
                                                                    if ((currentX-newI)/radius < -1.0) {
                                                                        startangle = acos(-1.0);
                                                                        } else {
                                                                        startangle = acos((currentX-newI)/radius);
                                                                        }
                                                                    }
                                                                if (currentY < newJ) startangle = 2 * M_PI - startangle;
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((newX - newI)/radius > 1.0) {
                                                                    endangle = acos(1.0);
                                                                    } else {
                                                                    if ((newX - newI)/radius < -1.0) {
                                                                        endangle = acos(-1.0);
                                                                        } else {
                                                                        endangle = acos((newX - newI)/radius);
                                                                        }
                                                                    }
                                                                if (newY < newJ) endangle = 2 * M_PI - endangle;
                                                                if (Arcdir == TCCW) {
                                                                    while (endangle <= startangle) {
                                                                        endangle += 2 * M_PI;
                                                                        }
                                                                    arclength = (endangle - startangle) * radius;
                                                                    stepangle = endangle - startangle;
                                                                    } else {
                                                                    while (endangle >= startangle) {
                                                                        endangle -= 2* M_PI;
                                                                        }
                                                                    arclength = (startangle - endangle) * radius /*/ (2.0 * M_PI)*/;
                                                                    stepangle = startangle - endangle;
                                                                    }
                                                                if (arclength > 0.0) {
                                                                    numberofsteps = (int) ceil(arclength/targetarclength);
                                                                    } else {
                                                                    numberofsteps = (int) ceil(-arclength/targetarclength);
                                                                    }
                                                                if (stepangle > 0.0) {
                                                                    numberofarcsteps = (int) ceil(stepangle/maxangle);
                                                                    } else {
                                                                    numberofarcsteps = (int) ceil(-stepangle/maxangle);
                                                                    }
                                                                if (numberofsteps < numberofarcsteps) numberofsteps = numberofarcsteps;

                                                                if (numberofsteps > 0) {
                                                                    injectedcounter = 0;
                                                                    //OK, now we have the number of steps  for the loop, now to find the other data
                                                                    double stepZ = (newZ - currentZ)/numberofsteps;
                                                                    double stepA = (newA - currentA)/numberofsteps;
                                                                    double stepB = (newB - currentB)/numberofsteps;
                                                                    double stepC = (newC - currentC)/numberofsteps;
                                                                    stepangle = (endangle - startangle)/numberofsteps;
                                                                    double currentangle = startangle;
                                                                    if (feedrateset) tempout = tempout + " F" + AnsiString(feedrate) + " \n";
                                                                    for (int i = 0; i < numberofsteps; i++) {
                                                                        localcounter++;
                                                                        currentangle += stepangle;
                                                                        if (relativeTranslation) {
                                                                            double oldX = newI + radius * cos(currentangle - stepangle);
                                                                            double oldY = newJ + radius * sin(currentangle - stepangle);
                                                                            newX = newI + radius * cos(currentangle);
                                                                            newY = newJ + radius * sin(currentangle);
                                                                            newZ = currentZ + ((double)(i + 1)) * stepZ;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            localtemp = "G1 X" + DP(4,newX - oldX) + " Y" + DP(4,newY - oldY);
                                                                            if (stepZ != 0.0) localtemp = localtemp + " Z" + DP(4,stepZ);
                                                                            if (stepA != 0.0) localtemp = localtemp + " A" + DP(4,stepA);
                                                                            if (stepB != 0.0) localtemp = localtemp + " B" + DP(4,stepB);
                                                                            if (stepC != 0.0) localtemp = localtemp + " C" + DP(4,stepC);
                                                                            if (G53found) localtemp = localtemp + " G53";
                                                                            } else {
                                                                            newX = newI + radius * cos(currentangle);
                                                                            newY = newJ + radius * sin(currentangle);
                                                                            newZ = currentZ + ((double)(i + 1)) * stepZ;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            localtemp = "G1 X" + DP(4,newX) + " Y" + DP(4,newY);
                                                                            if (stepZ != 0.0) localtemp = localtemp + " Z" + DP(4,newZ);
                                                                            if (stepA != 0.0) localtemp = localtemp + " A" + DP(4,newA);
                                                                            if (stepB != 0.0) localtemp = localtemp + " B" + DP(4,newB);
                                                                            if (stepC != 0.0) localtemp = localtemp + " C" + DP(4,newC);
                                                                            if (G53found) localtemp = localtemp + " G53";
                                                                            }
                                                                        //localtemp = localtemp + "\n";
                                                                        GCodeInputFile->Insert(localcounter,localtemp);
                                                                        injectedcounter++;
                                                                        //next line for debuggery only
                                                                        //GCodeOutputFile->Add("(Injected " + localtemp + ")");
                                                                        }
                                                                    //do not update these positions, as the text is injected back into the input filestream as a straight line rather than going to the output
                                                                    //injectedcounter keeps track of the number of lines injected as these will definitely be in mm, while the rest may be in inches.
                                                                    //currentX = newX;
                                                                    //currentY = newY;
                                                                    //currentZ = newZ;
                                                                    //currentA = newA;
                                                                    //currentB = newB;
                                                                    //currentC = newC;
                                                                    //if (stepZ != 0.0) currentZ = newZ;
                                                                    if (firstmove) {
                                                                        firstmove = false;
                                                                        }
                                                                    //do not update time, as these moves have been injected back into the inputfile stream
                                                                    //currentTime += arclength * 60.0/feedrate;
                                                                    } else {
                                                                    unknowncode = true;
                                                                    tempout = tempout + "\n(G" + AnsiString(codeno) + " Gives zero arc size)\n";
                                                                    Errors.Length++;
                                                                    Errors[Errors.Length - 1] = TZeroSizeArc;
                                                                    }
                                                                }
                                                                break;
                                                case TXZPlane:  {
                                                                //right, the arc is in the XZ plane
                                                                //This is included so the tool tip gets to the correct place, but does not write to the output file
                                                                double radius;
                                                                double startangle;
                                                                double endangle;
                                                                double arclength;
                                                                double stepangle;
                                                                int numberofsteps;
                                                                int numberofarcsteps;
                                                                radius = sqrt(pow(currentX - newI,2) + pow(currentZ - newK,2));
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((currentX-newI)/radius > 1.0) {
                                                                    startangle = acos(1.0);
                                                                    } else {
                                                                    if ((currentX-newI)/radius < -1.0) {
                                                                        startangle = acos(-1.0);
                                                                        } else {
                                                                        startangle = acos((currentX-newI)/radius);
                                                                        }
                                                                    }
                                                                if (currentZ < newK) startangle = 2 * M_PI - startangle;
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((newX - newI)/radius > 1.0) {
                                                                    endangle = acos(1.0);
                                                                    } else {
                                                                    if ((newX - newI)/radius < -1.0) {
                                                                        endangle = acos(-1.0);
                                                                        } else {
                                                                        endangle = acos((newX - newI)/radius);
                                                                        }
                                                                    }
                                                                if (newZ < newK) endangle = 2 * M_PI - endangle;
                                                                if (Arcdir == TCCW) {
                                                                    while (endangle < startangle) {
                                                                        endangle += 2 * M_PI;
                                                                        }
                                                                    arclength = (endangle - startangle) * radius;
                                                                    stepangle = endangle - startangle;
                                                                    } else {
                                                                    while (endangle > startangle) {
                                                                        endangle -= 2* M_PI;
                                                                        }
                                                                    arclength = (startangle - endangle) * radius /*/ (2.0 * M_PI)*/;
                                                                    stepangle = startangle - endangle;
                                                                    }
                                                                if (arclength > 0.0) {
                                                                    numberofsteps = (int) ceil(arclength/targetarclength);
                                                                    } else {
                                                                    numberofsteps = (int) ceil(-arclength/targetarclength);
                                                                    }
                                                                if (stepangle > 0.0) {
                                                                    numberofarcsteps = (int) ceil(stepangle/maxangle);
                                                                    } else {
                                                                    numberofarcsteps = (int) ceil(-stepangle/maxangle);
                                                                    }
                                                                if (numberofsteps < numberofarcsteps) numberofsteps = numberofarcsteps;

                                                                if (numberofsteps > 0) {
                                                                    //OK, now we have the number of steps  for the loop, now to find the other data
                                                                    double stepY = (newY - currentY)/numberofsteps;
                                                                    double stepA = (newA - currentA)/numberofsteps;
                                                                    double stepB = (newB - currentB)/numberofsteps;
                                                                    double stepC = (newC - currentC)/numberofsteps;
                                                                    stepangle = (endangle - startangle)/numberofsteps;
                                                                    double currentangle = startangle;
                                                                    //if (feedrateset) tempout = tempout + " F" + AnsiString(feedrate) + " \n";
                                                                    for (int i = 0; i < numberofsteps; i++) {
                                                                        currentangle += stepangle;
                                                                        if (relativeTranslation) {
                                                                            double oldX = newI + radius * cos(currentangle - stepangle);
                                                                            double oldZ = newK + radius * sin(currentangle - stepangle);
                                                                            newX = newI + radius * cos(currentangle);
                                                                            newZ = newK + radius * sin(currentangle);
                                                                            newY = currentY + ((double)(i + 1)) * stepY;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            //tempout = tempout + "G1 X" + DP(4,newX - oldX);
                                                                            //if (stepY != 0.0) tempout = tempout + " Y" + DP(4,stepY);
                                                                            //tempout = tempout + " Z" + DP(4,newZ - oldZ);
                                                                            //if (stepA != 0.0) tempout = tempout + " A" + DP(4,stepA);
                                                                            //if (stepB != 0.0) tempout = tempout + " B" + DP(4,stepB);
                                                                            //if (stepC != 0.0) tempout = tempout + " C" + DP(4,stepC);
                                                                            //if (G53found) tempout = tempout + " G53";
                                                                            } else {
                                                                            newX = newI + radius * cos(currentangle);
                                                                            newZ = newK + radius * sin(currentangle);
                                                                            newY = currentY + ((double)(i + 1)) * stepY;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            //tempout = tempout + "G1 X" + DP(4,newX);
                                                                            //if (stepY != 0.0) tempout = tempout + " Y" + DP(4,newY);
                                                                            //tempout = tempout + " Z" + DP(4,newZ);
                                                                            //if (stepA != 0.0) tempout = tempout + " A" + DP(4,newA);
                                                                            //if (stepB != 0.0) tempout = tempout + " B" + DP(4,newB);
                                                                            //if (stepC != 0.0) tempout = tempout + " C" + DP(4,newC);
                                                                            //if (G53found) tempout = tempout + " G53";
                                                                            }
                                                                        //tempout = tempout + "\n";
                                                                        }
                                                                    currentX = newX;
                                                                    currentY = newY;
                                                                    currentZ = newZ;
                                                                    currentA = newA;
                                                                    currentB = newB;
                                                                    currentC = newC;
                                                                    if (stepY != 0.0) currentY = newY;
                                                                    if (firstmove) {
                                                                        firstmove = false;
                                                                        }
                                                                    currentTime += arclength * 60.0/feedrate;
                                                                    } else {
                                                                    unknowncode = true;
                                                                    tempout = tempout + "\n(G" + AnsiString(codeno) + " Gives zero arc size)\n";
                                                                    Errors.Length++;
                                                                    Errors[Errors.Length - 1] = TZeroSizeArc;
                                                                    }
                                                                }
                                                                break;
                                                case TYZPlane:  {
                                                                //right, the arc is in the YZ plane
                                                                //This is included so the tool tip gets to the correct place, but does not write to the output file
                                                                double radius;
                                                                double startangle;
                                                                double endangle;
                                                                double arclength;
                                                                double stepangle;
                                                                int numberofsteps;
                                                                int numberofarcsteps;
                                                                radius = sqrt(pow(currentY - newJ,2) + pow(currentZ - newK,2));
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((currentY-newJ)/radius > 1.0) {
                                                                    startangle = acos(1.0);
                                                                    } else {
                                                                    if ((currentY-newJ)/radius < -1.0) {
                                                                        startangle = acos(-1.0);
                                                                        } else {
                                                                        startangle = acos((currentY-newJ)/radius);
                                                                        }
                                                                    }
                                                                if (currentZ < newK) startangle = 2 * M_PI - startangle;
                                                                //preventing acos domain errors due to rounding errors
                                                                if ((newY - newJ)/radius > 1.0) {
                                                                    endangle = acos(1.0);
                                                                    } else {
                                                                    if ((newY - newJ)/radius < -1.0) {
                                                                        endangle = acos(-1.0);
                                                                        } else {
                                                                        endangle = acos((newY - newJ)/radius);
                                                                        }
                                                                    }
                                                                if (newZ < newK) endangle = 2 * M_PI - endangle;
                                                                if (Arcdir == TCCW) {
                                                                    while (endangle < startangle) {
                                                                        endangle += 2 * M_PI;
                                                                        }
                                                                    arclength = (endangle - startangle) * radius;
                                                                    stepangle = endangle - startangle;
                                                                    } else {
                                                                    while (endangle > startangle) {
                                                                        endangle -= 2* M_PI;
                                                                        }
                                                                    arclength = (startangle - endangle) * radius /*/ (2.0 * M_PI)*/;
                                                                    stepangle = startangle - endangle;
                                                                    }
                                                                if (arclength > 0.0) {
                                                                    numberofsteps = (int) ceil(arclength/targetarclength);
                                                                    } else {
                                                                    numberofsteps = (int) ceil(-arclength/targetarclength);
                                                                    }
                                                                if (stepangle > 0.0) {
                                                                    numberofarcsteps = (int) ceil(stepangle/maxangle);
                                                                    } else {
                                                                    numberofarcsteps = (int) ceil(-stepangle/maxangle);
                                                                    }
                                                                if (numberofsteps < numberofarcsteps) numberofsteps = numberofarcsteps;

                                                                if (numberofsteps > 0) {
                                                                    //OK, now we have the number of steps  for the loop, now to find the other data
                                                                    double stepX = (newX - currentX)/numberofsteps;
                                                                    double stepA = (newA - currentA)/numberofsteps;
                                                                    double stepB = (newB - currentB)/numberofsteps;
                                                                    double stepC = (newC - currentC)/numberofsteps;
                                                                    stepangle = (endangle - startangle)/numberofsteps;
                                                                    double currentangle = startangle;
                                                                    //if (feedrateset) tempout = tempout + " F" + AnsiString(feedrate) + " \n";
                                                                    for (int i = 0; i < numberofsteps; i++) {
                                                                        currentangle += stepangle;
                                                                        if (relativeTranslation) {
                                                                            double oldY = newJ + radius * cos(currentangle - stepangle);
                                                                            double oldZ = newK + radius * sin(currentangle - stepangle);
                                                                            newY = newJ + radius * cos(currentangle);
                                                                            newZ = newK + radius * sin(currentangle);
                                                                            newX = currentX + ((double)(i + 1)) * stepX;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            //tempout = tempout + "G1";
                                                                            //if (stepX != 0.0) tempout = tempout + " X" + DP(4,stepX);
                                                                            //tempout = tempout + " Y" + DP(4,newY - oldY);
                                                                            //tempout = tempout + " Z" + DP(4,newZ - oldZ);
                                                                            //if (stepA != 0.0) tempout = tempout + " A" + DP(4,stepA);
                                                                            //if (stepB != 0.0) tempout = tempout + " B" + DP(4,stepB);
                                                                            //if (stepC != 0.0) tempout = tempout + " C" + DP(4,stepC);
                                                                            //if (G53found) tempout = tempout + " G53";
                                                                            } else {
                                                                            newY = newJ + radius * cos(currentangle);
                                                                            newZ = newK + radius * sin(currentangle);
                                                                            newX = currentX + ((double)(i + 1)) * stepX;
                                                                            newA = currentA + ((double)(i + 1)) * stepA;
                                                                            newB = currentB + ((double)(i + 1)) * stepB;
                                                                            newC = currentC + ((double)(i + 1)) * stepC;
                                                                            //tempout = tempout + "G1";
                                                                            //if (stepX != 0.0) tempout = tempout + " X" + DP(4,newX);
                                                                            //tempout = tempout + " Y" + DP(4,newY);
                                                                            //tempout = tempout + " Z" + DP(4,newZ);
                                                                            //if (stepA != 0.0) tempout = tempout + " A" + DP(4,newA);
                                                                            //if (stepB != 0.0) tempout = tempout + " B" + DP(4,newB);
                                                                            //if (stepC != 0.0) tempout = tempout + " C" + DP(4,newC);
                                                                            //if (G53found) tempout = tempout + " G53";
                                                                            }
                                                                        //tempout = tempout + "\n";
                                                                        }
                                                                    currentX = newX;
                                                                    currentY = newY;
                                                                    currentZ = newZ;
                                                                    currentA = newA;
                                                                    currentB = newB;
                                                                    currentC = newC;
                                                                    if (stepX != 0.0) currentX = newX;
                                                                    if (firstmove) {
                                                                        firstmove = false;
                                                                        }
                                                                    currentTime += arclength * 60.0/feedrate;
                                                                    } else {
                                                                    unknowncode = true;
                                                                    tempout = tempout + "\n(G" + AnsiString(codeno) + " Gives zero arc size)\n";
                                                                    Errors.Length++;
                                                                    Errors[Errors.Length - 1] = TZeroSizeArc;
                                                                    }
                                                                }
                                                                break;
                                                default:        {
                                                                unknowncode = true;
                                                                tempout = tempout + "\n(G" + AnsiString(codeno) + ")\n ";
                                                                Errors.Length++;
                                                                Errors[Errors.Length - 1] = TPlaneType;
                                                                }
                                                }
                                            if (feedrateset) currentFeedrate = feedrate;

                                            } //end of G2/G3
                                            break;
                                case  4:    {
                                            //Pause
                                            if (currentPlane == TXYPlane) tempout = tempout + "G4 ";
                                            }
                                            break;
                                case 17:    {
                                            //XY plane
                                            currentPlane = TXYPlane;
                                            tempout = tempout + "G17 ";
                                            }
                                            break;
                                case 18:    {
                                            //XZ plane
                                            currentPlane = TXZPlane;
                                            if (currentPlane == TXYPlane) tempout = tempout + "G18 ";
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TXZPresent;
                                            }
                                            break;
                                case 19:    {
                                            //YZ plane
                                            currentPlane = TYZPlane;
                                            if (currentPlane == TXYPlane) tempout = tempout + "G19 ";
                                            Errors.Length++;
                                            Errors[Errors.Length - 1] = TYZPresent;
                                            }
                                            break;
                                case 70:
                                case 20:    {
                                            //inch units. This is the 21st Century not the 19th Century, I can't be doing with inches, convert to metric and be done with the silly units.
                                            Units = TInch;
                                            tempout = tempout + "G21 ";
                                            }
                                            break;
                                case 71:
                                case 21:    {
                                            //metric units.
                                            Units = TMetric;
                                            tempout = tempout + "G21 ";
                                            }
                                            break;
                                case 40:    {
                                            //Cutter Compensation off.
                                            cutterCompensation = TCCNone;
                                            if (currentPlane == TXYPlane) tempout = tempout + "G40 ";
                                            }
                                            break;
                                case 42:
                                case 41:    {
                                            //Cutter Compensation left or right.
                                            if (codeno == 41.0) {
                                                //left compensation
                                                cutterCompensation = TCCLeft;
                                                tempout = tempout + "G41 ";
                                                } else {
                                                //right compensation
                                                cutterCompensation = TCCRight;
                                                tempout = tempout + "G42 ";
                                                }
                                            bool newDfound = false;
                                            int D;
                                            bool newPfound = false;
                                            double P;
                                            while (temp.Length() && (LeftString(temp,1) == "D" ||
                                                    LeftString(temp,1) == "P"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    double scale;
                                                    if (Units == TInch) {
                                                        scale = 25.4;
                                                        } else {
                                                        scale = 1.0;
                                                        }
                                                    if (localInstruction == "D") {
                                                        D = (int) localcodeno;
                                                        newDfound = true;
                                                        }
                                                    if (localInstruction == "P") {
                                                        P = localcodeno * scale;
                                                        newPfound = true;
                                                        }
                                                    }
                                            if (newDfound && currentPlane == TXYPlane) tempout = tempout + "D" + AnsiString(D) + " ";
                                            if (newPfound && currentPlane == TXYPlane) tempout = tempout + "P" + DP(4,P) + " ";
                                            }
                                            break;
                                case 44:
                                case 43:    {
                                            //Tool Length Compensation
                                            //don't need to store data with these
                                            if (codeno == 43.0) {
                                                //positive compensation
                                                tempout = tempout + "G43 ";
                                                } else {
                                                //negative compensation
                                                tempout = tempout + "G44 ";
                                                }
                                            bool newHfound = false;
                                            int H;
                                            while (temp.Length() && (LeftString(temp,1) == "H" )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "H") {
                                                        H = (int) localcodeno;
                                                        newHfound = true;
                                                        }
                                                    }
                                            if (newHfound && currentPlane == TXYPlane) tempout = tempout + "H" + AnsiString(H) + " ";
                                            }
                                            break;
                                case 49:    {
                                            //tool length compensation off
                                            if (currentPlane == TXYPlane) tempout = tempout + "G49 ";
                                            }
                                            break;
                                case 50:    {
                                            //scale factors turned to 1.0 in each axis
                                            if (currentPlane == TXYPlane) tempout = tempout + "G50 ";
                                            }
                                            break;
                                case 51:    {
                                            //scale factors specified for some or all axes
                                            //not going to keep a record of them in the program, as it doesn't seem very useful
                                            tempout = tempout + "G51 ";
                                            double scaleA = 1.0;
                                            double scaleB = 1.0;
                                            double scaleC = 1.0;
                                            double scaleX = 1.0;
                                            double scaleY = 1.0;
                                            double scaleZ = 1.0;
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "A") scaleA = localcodeno;
                                                    if (localInstruction == "B") scaleB = localcodeno;
                                                    if (localInstruction == "C") scaleC = localcodeno;
                                                    if (localInstruction == "X") scaleX = localcodeno;
                                                    if (localInstruction == "Y") scaleY = localcodeno;
                                                    if (localInstruction == "Z") scaleZ = localcodeno;
                                                    }
                                            if (scaleA != 1.0 && currentPlane == TXYPlane) tempout = tempout + "A" + DP(4,scaleA) + " ";
                                            if (scaleB != 1.0 && currentPlane == TXYPlane) tempout = tempout + "B" + DP(4,scaleB) + " ";
                                            if (scaleC != 1.0 && currentPlane == TXYPlane) tempout = tempout + "C" + DP(4,scaleC) + " ";
                                            if (scaleX != 1.0 && currentPlane == TXYPlane) tempout = tempout + "X" + DP(4,scaleX) + " ";
                                            if (scaleY != 1.0 && currentPlane == TXYPlane) tempout = tempout + "Y" + DP(4,scaleY) + " ";
                                            if (scaleZ != 1.0 && currentPlane == TXYPlane) tempout = tempout + "Z" + DP(4,scaleZ) + " ";
                                            }
                                            break;
                                case 92:
                                case 52:    {
                                            //offsets specified for some or all axes
                                            //not going to keep a record of them in the program, as it doesn't seem very useful
                                            tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            double offsetA = 0.0;
                                            double offsetB = 0.0;
                                            double offsetC = 0.0;
                                            double offsetX = 0.0;
                                            double offsetY = 0.0;
                                            double offsetZ = 0.0;
                                            bool newAfound = false;
                                            bool newBfound = false;
                                            bool newCfound = false;
                                            bool newXfound = false;
                                            bool newYfound = false;
                                            bool newZfound = false;
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "C" ||
                                                    LeftString(temp,1) == "X" ||
                                                    LeftString(temp,1) == "Y" ||
                                                    LeftString(temp,1) == "Z"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "A") {
                                                        offsetA = localcodeno;
                                                        newAfound = true;
                                                        }
                                                    if (localInstruction == "B") {
                                                        offsetB = localcodeno;
                                                        newBfound = true;
                                                        }
                                                    if (localInstruction == "C") {
                                                        offsetC = localcodeno;
                                                        newCfound = true;
                                                        }
                                                    if (localInstruction == "X") {
                                                        offsetX = localcodeno;
                                                        newXfound = true;
                                                        }
                                                    if (localInstruction == "Y") {
                                                        offsetY = localcodeno;
                                                        newYfound = true;
                                                        }
                                                    if (localInstruction == "Z") {
                                                        offsetZ = localcodeno;
                                                        newZfound = true;
                                                        }
                                                    }
                                            if (newAfound && currentPlane == TXYPlane) tempout = tempout + "A" + DP(4,offsetA) + " ";
                                            if (newBfound && currentPlane == TXYPlane) tempout = tempout + "B" + DP(4,offsetB) + " ";
                                            if (newCfound && currentPlane == TXYPlane) tempout = tempout + "C" + DP(4,offsetC) + " ";
                                            if (newXfound && currentPlane == TXYPlane) tempout = tempout + "X" + DP(4,offsetX) + " ";
                                            if (newYfound && currentPlane == TXYPlane) tempout = tempout + "Y" + DP(4,offsetY) + " ";
                                            if (newZfound && currentPlane == TXYPlane) tempout = tempout + "Z" + DP(4,offsetZ) + " ";
                                            }
                                            break;
                                case 58:
                                case 57:
                                case 56:
                                case 55:
                                case 54:    {
                                            //work offset
                                            if (currentPlane == TXYPlane) tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            }
                                            break;
                                case 59:    {
                                            //work offset with P qualifier
                                            tempout = tempout + "G59 ";
                                            bool newPfound = false;
                                            int P;
                                            while (temp.Length() && (LeftString(temp,1) == "P" )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "P") {
                                                        P = (int) localcodeno;
                                                        newPfound = true;
                                                        }
                                                    }
                                            if (newPfound && currentPlane == TXYPlane) tempout = tempout + "P" + AnsiString(P) + " ";
                                            }
                                            break;
                                case 61:    {
                                            //path control mode = exact stop
                                            pathControl = TExact;
                                            if (currentPlane == TXYPlane) tempout = tempout + "G61 ";
                                            }
                                case 64:    {
                                            //path control mode = exact stop
                                            pathControl = TVelocity;
                                            if (currentPlane == TXYPlane) tempout = tempout + "G64 ";
                                            }
                                case 68:    {
                                            //rotate co-ordinate system
                                            //not going to keep a record of them in the program, as it doesn't seem very useful
                                            //This causes an error in Mach 3, so won't add it to the output
                                            //tempout = tempout + "G68 ";
                                            double offsetA = 0.0;
                                            double offsetB = 0.0;
                                            double offsetR = 0.0;
                                            double offsetI = 0.0;
                                            bool newAfound = false;
                                            bool newBfound = false;
                                            bool newRfound = false;
                                            bool newIfound = false;
                                            //find out remainder of pertinant data in line
                                            while (temp.Length() && (LeftString(temp,1) == "A" ||
                                                    LeftString(temp,1) == "B" ||
                                                    LeftString(temp,1) == "R" ||
                                                    LeftString(temp,1) == "I"
                                                    )) {
                                                    AnsiString localInstruction = LeftString(temp,1);
                                                    temp = DeleteWord(temp,localInstruction);
                                                    temp = temp + " ";
                                                    double localcodeno = ToDoubleDef(ExtractWord(temp," "),0.0);
                                                    temp = DeleteWord(temp," ");
                                                    if (localInstruction == "A") {
                                                        offsetA = localcodeno;
                                                        newAfound = true;
                                                        }
                                                    if (localInstruction == "B") {
                                                        offsetB = localcodeno;
                                                        newBfound = true;
                                                        }
                                                    if (localInstruction == "R") {
                                                        offsetR = localcodeno;
                                                        newRfound = true;
                                                        }
                                                    if (localInstruction == "I") {
                                                        offsetI = localcodeno;
                                                        newIfound = true;
                                                        }
                                                    }
                                            //if (newAfound && currentPlane == TXYPlane) tempout = tempout + "A" + DP(4,offsetA) + " ";
                                            //if (newBfound && currentPlane == TXYPlane) tempout = tempout + "B" + DP(4,offsetB) + " ";
                                            //if (newIfound && currentPlane == TXYPlane) tempout = tempout + "I" + DP(4,offsetI) + " ";
                                            //if (newRfound && currentPlane == TXYPlane) tempout = tempout + "R" + DP(4,offsetR) + " ";
                                            }
                                            break;
                                case 69:    {
                                            //cancel co-ordinate rotation
                                            if (currentPlane == TXYPlane) tempout = tempout + "G69 ";
                                            }
                                case 90:    {
                                            //Absolute Distance. Now may be G90 or G90.1
                                            if (codeno == 90.1) {
                                                //circle centres done absolute
                                                relativeCentres = false;
                                                if (currentPlane == TXYPlane) tempout = tempout + "G90.1 ";
                                                } else {
                                                //translations done absolute
                                                relativeTranslation = false;
                                                if (currentPlane == TXYPlane) tempout = tempout + "G90 ";
                                                }
                                            }
                                            break;
                                case 91:    {
                                            //Absolute Distance. Now may be G91 or G91.1
                                            if (codeno == 91.1) {
                                                //circle centres done relative
                                                relativeCentres = true;
                                                if (currentPlane == TXYPlane) tempout = tempout + "G91.1 ";
                                                } else {
                                                //translations done relative
                                                relativeTranslation = false;
                                                if (currentPlane == TXYPlane) tempout = tempout + "G91 ";
                                                }
                                            }
                                            break;
                                case 94:
                                case 93:    {
                                            //feed rate definition, 93 = per time, 94 = per min
                                            if (currentPlane == TXYPlane) tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            }
                                            break;
                                case 99:
                                case 98:    {
                                            //canned cycle return, different meanings allocated dependant on codeno
                                            if (currentPlane == TXYPlane) tempout = tempout + "G" + AnsiString(codeno) + " ";
                                            }
                                            break;
                                default:    break;
                                }
                            }  //end of G codes

                        if (Instruction == "M") {
                            //it's an M Code
                            switch ((int)codeno) {
                                case 0:     {
                                            //Program Stop
                                            tempout = tempout + "M0 ";
                                            }
                                            break;
                                case 1:     {
                                            //Optional Program Stop
                                            tempout = tempout + "M1 ";
                                            }
                                            break;
                                case 2:     {
                                            //Program End
                                            tempout = tempout + "M2 ";
                                            }
                                            break;
                                case 3:     {
                                            //Rotate Spindle CW
                                            if (currentPlane == TXYPlane) tempout = tempout + "M3 ";
                                            currentSpindleDir = TCW;
                                            }
                                            break;
                                case 4:     {
                                            //Rotate Spindle CCW
                                            if (currentPlane == TXYPlane) tempout = tempout + "M4 ";
                                            currentSpindleDir = TCCW;
                                            }
                                            break;
                                case 5:     {
                                            //Stop Spindle
                                            tempout = tempout + "M5 ";
                                            currentSpindleDir = TStop;
                                            }
                                            break;
                                case 6:     {
                                            //Change tool
                                            if (currentPlane == TXYPlane) tempout = tempout + "M6 ";
                                            }
                                            break;
                                case 7:     {
                                            //Coolant Mist
                                            if (currentPlane == TXYPlane) tempout = tempout + "M7 ";
                                            currentCoolant = TMist;
                                            }
                                            break;
                                case 8:     {
                                            //Coolant Flood
                                            if (currentPlane == TXYPlane) tempout = tempout + "M8 ";
                                            currentCoolant = TFlood;
                                            }
                                            break;
                                case 9:     {
                                            //Coolant None
                                            if (currentPlane == TXYPlane) tempout = tempout + "M9 ";
                                            currentCoolant = TNone;
                                            }
                                            break;
                                case 30:    {
                                            //Program Stop & Rewind
                                            tempout = tempout + "M30 ";
                                            }
                                            break;
                                case 47:    {
                                            //Program Stop & Rewind & restart from line 1
                                            tempout = tempout + "M47 ";
                                            }
                                            break;
                                case 48:    {
                                            //Enable feed override
                                            if (currentPlane == TXYPlane) tempout = tempout + "M48 ";
                                            }
                                            break;
                                case 49:    {
                                            //Disable feed override
                                            if (currentPlane == TXYPlane) tempout = tempout + "M49 ";
                                            }
                                            break;
                                case 60:    {
                                            //Program Stop
                                            tempout = tempout + "M60 ";
                                            }
                                            break;
                                default:    {
                                            //un recognised code found
                                            unknowncode = true;
                                            tempout = tempout + "\n(M" + AnsiString(codeno) + ")\n ";
                                            }
                                }
                            } //end of M codes  
                        if (Instruction == "N") {
                            //it's a Line number
                            if (currentPlane == TXYPlane) tempout = tempout + "N" + AnsiString(codeno) + " ";
                            }
                        if (Instruction == "O") {
                            //it's a subroutine label number
                            if (currentPlane == TXYPlane) tempout = tempout + "O" + AnsiString(codeno) + " ";
                            }
                        if (Instruction == "P") {
                            //it's a pause, in milliseconds
                            if (currentPlane == TXYPlane) tempout = tempout + "P" + AnsiString(codeno) + " ";
                            currentTime += codeno/1000.0;
                            }
                        if (Instruction == "R") {
                            //Canned Cycle retract level
                            if (currentPlane == TXYPlane) tempout = tempout + "R" + AnsiString(codeno) + " ";
                            currentRetractLevel = codeno;
                            }
                        if (Instruction == "S") {
                            //it's spindle speed, rpm
                            if (currentPlane == TXYPlane) tempout = tempout + "S" + AnsiString(codeno) + " ";
                            }
                        if (Instruction == "T") {
                            //it's a tool selection number
                            currentTool = (int)codeno;
                            if (currentPlane == TXYPlane) tempout = tempout + "T" + AnsiString(currentTool) + " ";
                            }

                        } //end of 'if (Instruction == "(")'
                    } //end of 'while (temp.Length())'

                if ((tempout.Trim()).Length()) GCodeOutputFile->Add(tempout.Trim());
                //debug
                //GCodeOutputFile->Add("(X=" + DP(4,currentX) + " Y=" + DP(4,currentY) + " Z=" + DP(4,currentZ) + " A=" + DP(4,currentA) + " B=" + DP(4,currentB) + " C=" + DP(4,currentC) + ")");
                counter ++;
                } //end of 'while (counter < filelength)'

            //calculate how long the machining program takes
            tempout = "";
            double hours = (double) ((int)(currentTime/3600.0));
            currentTime -= hours * 3600.0;
            double minutes = (double) ((int)(currentTime/60.0));
            currentTime -= minutes * 60.0;

            if (hours > 0.0) {
                tempout = tempout + AnsiString(hours) + " Hours, ";
                }
            if (minutes > 0.0) {
                tempout = tempout + AnsiString(minutes) + " Minutes, ";
                }
            tempout = tempout + DP(2,currentTime) + " Seconds.";


            tempout = "(Total Running Time minimum = " + tempout + " )";
            GCodeOutputFile->Insert(5,tempout);

            //Finish by saving then clearing memory
            Label3->Caption = "Saving File";
            Refresh();
            /*
            //debug
            for (int i = 0; i < GCodeInputFile->Count; i++) {
                GCodeOutputFile->Add("(" + GCodeInputFile->Strings[i] + ")");
                }
            */

            GCodeOutputFile->SaveToFile(Edit2->Text);

            Label3->Caption = "Clearing Memory";
            Refresh();

            delete GCodeInputFile;
            delete GCodeOutputFile;

            if (!unknowncode) {
                Label3->Caption = "Finished";
                } else {
                Label3->Caption = "Unknown Codes encountered, please check file";
                AnsiString Error = Errors.Length > 1 ? "s were " : " was ";
                Error = "The following Error" + Error + "encountered:\n";
                for (int i = 1; i < Errors.Length; i++) {
                    switch (Errors[i]) {
                        case TXZPresent  :  {
                                            Error = Error + "\tG18: XZ Plane definition found.\n";
                                            }
                                            break;
                        case TYZPresent  :  {
                                            Error = Error + "\tG19: YZ Plane definition encountered.\n";
                                            }
                                            break;
                        case TZeroSizeArc:  {
                                            Error = Error + "\tZero size arc met.\n";
                                            }
                                            break;
                        case TPlaneType  :  {
                                            Error = Error + "\tUnknown Working Plane definition specified. Possible program internal error.\n";
                                            }
                                            break;
                        case TUnknownMCode: {
                                            Error = Error + "\tUnknown M Code Encountered.\n";
                                            }
                                            break;
                        case TUnknownGCode: {
                                            Error = Error + "\tUnknown G Code Encountered.\n";
                                            }
                                            break;
                        case TAAxisInstruction: {
                                            Error = Error + "\tThe input file contains A Axis Instructions.\n";
                                            }
                                            break;
                        case TBAxisInstruction: {
                                            Error = Error + "\tThe input file contains B Axis Instructions.\n";
                                            }
                                            break;
                        case TCAxisInstruction: {
                                            Error = Error + "\tThe input file contains C Axis Instructions.\n";
                                            }
                                            break;
                        default:            {
                                            Error = Error + "\tProgram internal error.\n";
                                            }
                                            break;
                        }
                    }
                Errors.Length = 0;
                }
            Refresh();
            }  //end of try  statement

        __finally {
            WrapButton->Enabled = true;
            OriginalGCodeButton->Enabled = true;
            NewGCodeButton->Enabled = true;
            HelpButton->Enabled = true;
            CloseButton->Enabled = true;
            Edit1->Enabled = true;
            Edit2->Enabled = true;
            Edit3->Enabled = true;
            Edit4->Enabled = true;
            Refresh();
            }

        } else {
        if (FileExists(Edit1->Text)) {
            Application->MessageBox("An existing converted file will get overwritten.\n\nPlease delete file or change converted file name","Warning", MB_OK | MB_ICONERROR);
            } else {
            Application->MessageBox("The input file does not exist.\n\nPlease check","Warning", MB_OK | MB_ICONERROR);
            }
        Label3->Caption = "Conversion Aborted";
        }
    
}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::HelpButtonClick(TObject *Sender)
{
    //Application->MessageBox("This program takes a G Code File, and wraps\nthe code around the X or Y axis.\n\nTo operate, select 2 file names, max segment & degree values for the\narcs to be converted to lines, then fill in the pertinant axis data\n and press the\n'Wrap' button.\n","Help on G Code Wrapper Program",MB_OK | MB_ICONINFORMATION);
    AnsiString Address = ExtractFilePath(Application->ExeName) + "\\G Code Wrapper Help.pdf";
    ShellExecute(Application->Handle,NULL,Address.c_str(),NULL,NULL,0);
    Address.~AnsiString();

    
}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::CloseButtonClick(TObject *Sender)
{
    Close();    
}
//---------------------------------------------------------------------------
void __fastcall TWrapperForm::Edit5Change(TObject *Sender)
{
    Edit7->Text = Edit5->Text;
}
//---------------------------------------------------------------------------

